from typing import Optional

from httpx import HTTPStatusError


class ApiError(Exception):
    pass


class ApiStatusError(ApiError):
    def __init__(self, message: Optional[str] = None, *, exc: HTTPStatusError):
        super().__init__(message or str(exc))
        self.request = exc.request
        self.response = exc.response


class SSEResponseParseError(ApiError):
    pass


class SSERequestFailedError(ApiError):
    pass


# TODO (s): this if for backwards compatibility and remove this eventually
GatewayApiError = ApiError
GatewayApiStatusError = ApiStatusError



from decimal import Decimal
from enum import Enum
from typing import Optional

import attrs

from grazie.api.client.chat.prompt import ChatPrompt
from grazie.api.client.utils import typ, typ_or_none


@attrs.define(auto_attribs=True)
class LLMChatEventV5:
    type: str
    content: str
    name: Optional[str] = None


class LLMChatEventType(str, Enum):
    Content = "Content"
    FunctionCall = "FunctionCall"
    QuotaMetadata = "QuotaMetadata"


@attrs.define(auto_attribs=True)
class Credit:
    amount: Decimal


@attrs.define(auto_attribs=True)
class QuotaID:
    quotaId: str


@attrs.define(auto_attribs=True)
class Quota:
    license: str  # Deprecated
    until: int
    current: Credit = typ(Credit)
    maximum: Credit = typ(Credit)
    quotaID: QuotaID = typ(QuotaID)


@attrs.define(auto_attribs=True)
class LLMChatEventV6:
    type: LLMChatEventType = attrs.field(
        converter=LLMChatEventType  # pyright: ignore [reportGeneralTypeIssues]
    )
    content: str = ""
    name: Optional[str] = None
    updated: Optional[Quota] = typ_or_none(Quota)
    spent: Optional[Credit] = typ_or_none(Credit)


@attrs.define(auto_attribs=True, frozen=True)
class ChatResponse:
    prompt: ChatPrompt
    content: str
    function_call: Optional[str] = None
    updated: Optional[Quota] = None
    spent: Optional[Credit] = None


@attrs.define(auto_attribs=True, frozen=True)
class ChatResponseStream:
    chunk: str
    function_call: Optional[str] = None
    updated: Optional[Quota] = None
    spent: Optional[Credit] = None



from typing import Any, Dict, List, Optional

import attr
import attrs

from grazie.api.client.utils import typ


@attrs.define(auto_attribs=True, frozen=True)
class AnswerChunk:
    documents: Optional[List[Dict[str, Any]]] = None
    summaryChunk: Optional[str] = None
    summaryChunkType: Optional[str] = None
    summaryReferences: Optional[List[str]] = None


@attrs.define(auto_attribs=True, frozen=True)
class AnswerStreamV2:
    chunk: AnswerChunk = typ(AnswerChunk)


@attrs.define(auto_attribs=True, frozen=True)
class RetrieveResponse:
    documents: List[Dict[str, Any]]



#!/usr/bin/env python3

import argparse
import os
from contextlib import nullcontext

from grazie.api.client.chat.prompt import ChatPrompt
from grazie.api.client.completion.prompt import CompletionPrompt
from grazie.api.client.endpoints import GrazieApiGatewayUrls
from grazie.api.client.gateway import AuthType, GrazieAgent, GrazieApiGatewayClient
from grazie.api.client.profiles import LLMProfile, Profile


def get_client(args: argparse.Namespace) -> GrazieApiGatewayClient:
    return GrazieApiGatewayClient(
        grazie_agent=GrazieAgent(name="grazie-api-gateway-client", version="dev"),
        url=args.gateway,
        auth_type=AuthType.USER,
        grazie_jwt_token=args.token,
    )


def completion(args: argparse.Namespace) -> None:
    client = get_client(args)

    response = client.complete(
        prompt=CompletionPrompt(args.prompt),
        profile=Profile.get_by_name(args.profile),
    )

    print(response.completion)


def chat(args: argparse.Namespace) -> None:
    client = get_client(args)

    response = client.chat(
        chat=ChatPrompt().add_system(args.system).add_user(args.prompt),
        profile=Profile.get_by_name(args.profile),
    )

    print(response.content)


def main() -> None:
    profiles = sorted(
        [
            value.name
            for attr in dir(Profile)
            for value in (getattr(Profile, attr),)
            if isinstance(value, LLMProfile)
        ]
    )

    parser = argparse.ArgumentParser(description="Grazie API Gateway Client")
    parser.add_argument("-g", "--gateway", type=str, default=GrazieApiGatewayUrls.STAGING)
    parser.add_argument("-p", "--profile", choices=profiles, default=Profile.OPENAI_CHAT_GPT.name)
    parser.add_argument(
        "-t",
        "--token",
        type=str,
        default=os.environ.get("GRAZIE_JWT_TOKEN", None),
        help=argparse.SUPPRESS,
    )
    subparsers = parser.add_subparsers()

    with nullcontext(subparsers.add_parser("completion")) as subparser:
        subparser.add_argument("prompt")
        subparser.set_defaults(func=completion)

    with nullcontext(subparsers.add_parser("chat")) as subparser:
        subparser.add_argument("-s", "--system", type=str, default="You are a helpful assistant.")
        subparser.add_argument("prompt")
        subparser.set_defaults(func=chat)

    args = parser.parse_args()

    if hasattr(args, "func"):
        args.func(args)
    else:
        parser.print_help()


if __name__ == "__main__":
    main()



import json
from typing import Optional, Union

import httpx

from . import types
from ._client import AsyncClient, Client
from ._version import __version__
from .tasks.api import AsyncTasksAPI, TasksAPI

AGENT = json.dumps(dict(name="api-gateway-python-client", version=__version__))


class APIGatewayClient:
    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        auth_type: types.AuthType = types.AuthType.USER,
        endpoint: Union[str, types.GatewayEndpoint] = types.GatewayEndpoint.PRODUCTION,
        http_client: Optional[httpx.Client] = None,
    ):
        """Construct a new synchronous JetBrains AI api gateway client instance. It's a facade that combines all the APIs clients.

        It automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` and `auth_type` from `GRAZIE_JWT_TOKEN`, `GRAZIE_USER_JWT_TOKEN`, `GRAZIE_SERVICE_JWT_TOKEN` or `GRAZIE_APPLICATION_JWT_TOKEN`
        """
        self.client = Client(
            api_key=api_key,
            auth_type=auth_type,
            endpoint=endpoint,
            http_client=http_client,
            headers={
                "Grazie-Agent": AGENT,
            },
        )
        self.tasks = TasksAPI(self.client)


class AsyncAPIGatewayClient:
    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        auth_type: types.AuthType = types.AuthType.USER,
        endpoint: Union[str, types.GatewayEndpoint] = types.GatewayEndpoint.PRODUCTION,
        http_client: Optional[httpx.AsyncClient] = None,
    ):
        """Construct a new asynchronous JetBrains AI api gateway client instance. It's a facade that combines all the APIs clients.

        It automatically infers the following arguments from their corresponding environment variables if they are not provided:
        - `api_key` and `auth_type` from `GRAZIE_JWT_TOKEN`, `GRAZIE_USER_JWT_TOKEN`, `GRAZIE_SERVICE_JWT_TOKEN` or `GRAZIE_APPLICATION_JWT_TOKEN`
        """
        self.client = AsyncClient(
            api_key=api_key,
            auth_type=auth_type,
            endpoint=endpoint,
            http_client=http_client,
            headers={
                types.GrazieHeaders.AGENT: AGENT,
            },
        )
        self.tasks = AsyncTasksAPI(self.client)



from typing import Dict, List, Optional

import attr

from grazie.api.client.profiles import LLMProfile


@attr.s(auto_attribs=True)
class EmbeddingRequest:
    texts: List[str]
    model: Optional[str] = None
    normalize: bool = False
    format_cbor: bool = False


@attr.s(auto_attribs=True)
class LLMEmbeddingRequest:
    texts: List[str]
    profile: LLMProfile
    dimensions: Optional[int] = None



import abc
import json
from typing import Any, Dict, List

import attrs

from grazie.api.client.llm_functions import FunctionDefinition


class Parameters:
    @attrs.define(auto_attribs=True, frozen=True)
    class Key:
        fqdn: str
        type: str

        def serialize(self) -> Dict[str, Any]:
            return {"type": self.type, "fqdn": self.fqdn}

    @attrs.define(auto_attribs=True, frozen=True)
    class BooleanKey(Key):
        type: str = "bool"

    @attrs.define(auto_attribs=True, frozen=True)
    class IntKey(Key):
        type: str = "int"

    @attrs.define(auto_attribs=True, frozen=True)
    class FloatKey(Key):
        type: str = "double"

    @attrs.define(auto_attribs=True, frozen=True)
    class StrKey(Key):
        type: str = "text"

    @attrs.define(auto_attribs=True, frozen=True)
    class JsonKey(Key):
        type: str = "json"

    @attrs.define(auto_attribs=True, frozen=True)
    class Value:
        @abc.abstractmethod
        def serialize(self) -> Dict[str, Any]:
            pass

    @attrs.define(auto_attribs=True, frozen=True)
    class BooleanValue(Value):
        value: bool

        def serialize(self) -> Dict[str, Any]:
            return {"type": "bool", "value": self.value, "modified": 0}

    @attrs.define(auto_attribs=True, frozen=True)
    class IntValue(Value):
        value: int

        def serialize(self) -> Dict[str, Any]:
            return {"type": "int", "value": self.value, "modified": 0}

    @attrs.define(auto_attribs=True, frozen=True)
    class FloatValue(Value):
        value: float

        def serialize(self) -> Dict[str, Any]:
            return {"type": "double", "value": self.value, "modified": 0}

    @attrs.define(auto_attribs=True, frozen=True)
    class StrValue(Value):
        value: str

        def serialize(self) -> Dict[str, Any]:
            return {"type": "text", "value": self.value, "modified": 0}

    @attrs.define(auto_attribs=True, frozen=True)
    class JsonValue(Value):
        value: Any

        def serialize(self) -> Dict[str, Any]:
            return {"type": "json", "value": json.dumps(self.value), "modified": 0}

        @classmethod
        def from_functions(cls, *funcs: FunctionDefinition):
            return cls(value=[attrs.asdict(func) for func in funcs])

    _allowed_type_mapping = {
        (BooleanKey, BooleanValue),
        (IntKey, IntValue),
        (FloatKey, FloatValue),
        (StrKey, StrValue),
        (JsonKey, JsonValue),
    }

    @classmethod
    def serialize(cls, parameters: Dict[Key, Value]) -> List[Dict[str, Any]]:
        parameters_data: List[Dict[str, Any]] = []
        for key, value in parameters.items():
            if not (key.__class__, value.__class__) in cls._allowed_type_mapping:
                raise ValueError(
                    f"Types for key and value should be the same. "
                    f"Got {key.__class__.__name__}:{value.__class__.__name__}"
                )
            parameters_data.extend(
                [
                    key.serialize(),
                    value.serialize(),
                ]
            )
        return parameters_data



from enum import Enum
from typing import Any, Dict, List, Optional

import attr
import attrs
import jsonschema


def _check_parameters_schema(obj, attribute, value):
    try:
        jsonschema.Draft202012Validator.check_schema(value)
    except jsonschema.SchemaError as e:
        raise ValueError(f"Invalid parameters schema for function {obj.name}") from e


@attr.define(auto_attribs=True, frozen=True)
class FunctionDefinition:
    name: str
    description: Optional[str] = None
    parameters: Dict = attr.field(
        factory=lambda: {"type": "object", "properties": {}},
        validator=_check_parameters_schema,
    )

    class FunctionParameterTypes(Enum):
        BOOLEAN = "boolean"
        INTEGER = "integer"
        FLOAT = "float"
        STRING = "string"
        ARRAY = "array"

    def add_argument(
        self,
        name: str,
        description: Optional[str] = None,
        _type: Optional[FunctionParameterTypes] = None,
        required: bool = False,
        enum: Optional[List[Any]] = None,
    ) -> "FunctionDefinition":
        parameters = self.parameters
        parameters["properties"][name] = {}
        if description:
            parameters["properties"][name]["description"] = description
        if _type:
            parameters["properties"][name]["type"] = _type.value
        if enum is not None:
            parameters["properties"][name]["enum"] = enum
        if required:
            if "required" not in parameters:
                parameters["required"] = []
            parameters["required"].append(name)
        return attrs.evolve(self, parameters=parameters)



from decimal import Decimal
from typing import Optional

import attrs

from grazie.api.client.utils import typ


@attrs.define(auto_attribs=True, frozen=True)
class QuotaCredit:
    amount: Decimal


@attrs.define(auto_attribs=True, frozen=True)
class QuotaId:
    quotaId: str
    userId: Optional[str] = None


@attrs.define(auto_attribs=True, frozen=True)
class Quota:
    license: str
    until: int
    current: QuotaCredit = typ(QuotaCredit)
    maximum: QuotaCredit = typ(QuotaCredit)
    quotaID: QuotaId = typ(QuotaId)



import json
import logging
import os
from typing import AsyncIterator, Dict, Iterator, Optional, Union
from urllib.parse import urljoin

import httpx

from grazie.api.client.gateway import AuthType as OldAuthType

from . import exceptions, types
from ._sse import _decode_sse_line

logger = logging.getLogger(__name__)


class ClientBase:
    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        auth_type: Union[types.AuthType, OldAuthType] = types.AuthType.USER,
        endpoint: Union[str, types.GatewayEndpoint] = types.GatewayEndpoint.PRODUCTION,
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ):
        if api_key is None:
            if (api_key := os.getenv("GRAZIE_JWT_TOKEN")) or (
                api_key := os.getenv("GRAZIE_USER_JWT_TOKEN")
            ):
                auth_type = types.AuthType.USER
            elif api_key := os.getenv("GRAZIE_SERVICE_JWT_TOKEN"):
                auth_type = types.AuthType.SERVICE
            elif api_key := os.getenv("GRAZIE_APPLICATION_JWT_TOKEN"):
                auth_type = types.AuthType.APPLICATION
            else:
                raise ValueError(
                    (
                        "Cannot set jwt token. Either pass it as a constructor param or set GRAZIE_JWT_TOKEN environment variable.\n"
                        "You can obtain user jwt token by going to https://try.ai.intellij.net and copying it by pressing a button in a top right corner."
                    )
                )

        self.api_key = api_key
        self.auth_type = types.AuthType(auth_type.value)
        self.endpoint = endpoint.value if isinstance(endpoint, types.GatewayEndpoint) else endpoint
        self.headers = headers or {}

    def _prepare_headers(self, headers: Optional[Dict[str, str]] = None) -> dict:
        return {
            types.GrazieHeaders.AUTH_TOKEN: self.api_key,
            **self.headers,
            **(headers or {}),
        }

    def _prepare_url(self, path: str) -> str:
        return urljoin(str(self.endpoint), f"{self.auth_type.path}/{path}")

    def _raise_for_status(self, response: httpx.Response):
        try:
            response.raise_for_status()
        except httpx.HTTPStatusError as e:
            raise exceptions.ApiStatusError(
                message=f"{e.response.status_code}: {e.response.text}", exc=e
            ) from e


class Client(ClientBase):
    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        auth_type: types.AuthType = types.AuthType.USER,
        endpoint: Union[str, types.GatewayEndpoint] = types.GatewayEndpoint.PRODUCTION,
        headers: Optional[Dict[str, str]] = None,
        http_client: Optional[httpx.Client] = None,
    ):
        """A wrapper around httpx.Client, with headers / auth / common request and stream handling logic."""
        super().__init__(api_key=api_key, auth_type=auth_type, endpoint=endpoint, headers=headers)
        self.http_client = http_client or httpx.Client()

    def stream(
        self,
        method: str,
        path: str,
        *,
        json=None,
        params=None,
        headers=None,
        raise_for_status: bool = True,
        **kwargs,
    ) -> Iterator[Dict[str, str]]:
        url = self._prepare_url(path)
        headers = self._prepare_headers(headers)

        with self.http_client.stream(
            method, url, json=json, params=params, headers=headers, **kwargs
        ) as response:
            if raise_for_status and response.is_error:
                # For streaming responses we need to first
                # read their content to correctly access response.text
                response.read()
                self._raise_for_status(response)

            for line in response.iter_lines():
                if decoded := _decode_sse_line(line):
                    yield decoded

    def request(
        self,
        method: str,
        path: str,
        *,
        json=None,
        params=None,
        headers=None,
        raise_for_status: bool = True,
        **kwargs,
    ) -> httpx.Response:
        url = self._prepare_url(path)
        headers = self._prepare_headers(headers)

        response = self.http_client.request(
            method,
            url,
            json=json,
            params=params,
            headers=headers,
            **kwargs,
        )
        if raise_for_status:
            self._raise_for_status(response)
        return response


class AsyncClient(ClientBase):
    def __init__(
        self,
        *,
        api_key: Optional[str] = None,
        auth_type: types.AuthType = types.AuthType.USER,
        endpoint: Union[str, types.GatewayEndpoint] = types.GatewayEndpoint.PRODUCTION,
        headers: Optional[Dict[str, str]] = None,
        http_client: Optional[httpx.AsyncClient] = None,
    ):
        """A wrapper around httpx.AsyncClient, with headers / auth / common request and stream handling logic."""
        super().__init__(api_key=api_key, auth_type=auth_type, endpoint=endpoint, headers=headers)
        self.http_client = http_client or httpx.AsyncClient()

    async def stream(
        self,
        method: str,
        path: str,
        *,
        json=None,
        params=None,
        headers=None,
        raise_for_status: bool = True,
        **kwargs,
    ) -> AsyncIterator[Dict[str, str]]:
        url = self._prepare_url(path)
        headers = self._prepare_headers(headers)

        async with self.http_client.stream(
            method, url, json=json, params=params, headers=headers, **kwargs
        ) as response:
            if raise_for_status and response.is_error:
                # For streaming responses we need to first
                # read their content to correctly access response.text
                await response.aread()
                self._raise_for_status(response)
            async for line in response.aiter_lines():
                if decoded := _decode_sse_line(line):
                    yield decoded

    async def request(
        self,
        method: str,
        path: str,
        *,
        json=None,
        params=None,
        headers=None,
        raise_for_status: bool = True,
        **kwargs,
    ) -> httpx.Response:
        url = self._prepare_url(path)
        headers = self._prepare_headers(headers)

        response = await self.http_client.request(
            method,
            url,
            json=json,
            params=params,
            headers=headers,
            **kwargs,
        )
        if raise_for_status:
            self._raise_for_status(response)
        return response



from typing import Optional

import attr
import attrs

from grazie.api.client.chat.roles import ChatRole


@attrs.define(auto_attribs=True, frozen=True)
class ChatMessage:
    def to_json(self) -> dict:
        raise NotImplementedError


@attrs.define(auto_attribs=True, frozen=True)
class TextChatMessage(ChatMessage):
    """
    A message that is just text.
    """

    role: ChatRole
    text: str

    def to_json(self) -> dict:
        return {"type": self.role.value.lower() + "_message", "content": self.text}


@attrs.define(auto_attribs=True, frozen=True)
class AssistantChatMessage(ChatMessage):
    """
    Assistant LLM chat message.
    Used for model-generated messages including model-generated functionCall response.

    :param text - for common model-generated messages
    :param functionCall - for model-generated functionCall response
    """

    @attrs.define(auto_attribs=True, frozen=True)
    class FunctionCallResponse:
        """
        :param functionName - the name of the function which was called
        :param content - the content of the function call response
        """

        functionName: str
        content: str

    text: Optional[str]
    functionCall: Optional[FunctionCallResponse] = None

    def to_json(self) -> dict:
        return {
            "type": ChatRole.ASSISTANT.value.lower() + "_message",
            "content": self.text,
            "functionCall": (attr.asdict(self.functionCall) if self.functionCall else None),
        }


@attrs.define(auto_attribs=True, frozen=True)
class FunctionCallChatMessage(ChatMessage):
    """
    This message is used to specify a result of a function execution with model-generated parameters.
    """

    name: str
    content: str

    def to_json(self) -> dict:
        return {
            "type": ChatRole.FUNCTION.value.lower() + "_message",
            "content": self.content,
            "functionName": self.name,
        }



from ._client import AsyncClient, Client
from ._gateway_client import APIGatewayClient, AsyncAPIGatewayClient
from .types import AuthType, GatewayEndpoint

__all__ = [
    "APIGatewayClient",
    "AsyncAPIGatewayClient",
    "AuthType",
    "Client",
    "AsyncClient",
    "GatewayEndpoint",
]



import attrs


@attrs.define(auto_attribs=True, frozen=True)
class LLMProfile:
    name: str

    def organization(self) -> str:
        return self.name.split("-", maxsplit=1)[0]


@attrs.define(auto_attribs=True, frozen=True)
class GrazieGptNeoTinyTextProfile(LLMProfile):
    name: str = "grazie-gpt-neo-tiny-text"


@attrs.define(auto_attribs=True, frozen=True)
class GrazieReplitCodeV1SmallProfile(LLMProfile):
    name: str = "grazie-replit-code-v1-small"


@attrs.define(auto_attribs=True, frozen=True)
class GrazieBigCodeStarcoderProfile(LLMProfile):
    name: str = "grazie-bigcode-starcoder"


@attrs.define(auto_attribs=True, frozen=True)
class GrazieChatLlamaV2_7bProfile(LLMProfile):
    name: str = "grazie-chat-llama-v2-7b"


@attrs.define(auto_attribs=True, frozen=True)
class GrazieChatLlamaV2_13bProfile(LLMProfile):
    name: str = "grazie-chat-llama-v2-13b"


@attrs.define(auto_attribs=True, frozen=True)
class GrazieCodeLlama7bProfile(LLMProfile):
    name: str = "grazie-code-llama-7b"


@attrs.define(auto_attribs=True, frozen=True)
class GrazieCodeLlama13bProfile(LLMProfile):
    name: str = "grazie-code-llama-13b"


@attrs.define(auto_attribs=True, frozen=True)
class AnthropicClaudeProfile(LLMProfile):
    name: str = "anthropic-claude"


@attrs.define(auto_attribs=True, frozen=True)
class AnthropicClaudeInstantProfile(LLMProfile):
    name: str = "anthropic-claude-instant"


@attrs.define(auto_attribs=True, frozen=True)
class AnthropicClaude3HaikuProfile(LLMProfile):
    name: str = "anthropic-claude-3-haiku"


@attrs.define(auto_attribs=True, frozen=True)
class AnthropicClaude3SonnetProfile(LLMProfile):
    name: str = "anthropic-claude-3-sonnet"


@attrs.define(auto_attribs=True, frozen=True)
class AnthropicClaude3OpusProfile(LLMProfile):
    name: str = "anthropic-claude-3-opus"


@attrs.define(auto_attribs=True, frozen=True)
class AnthropicClaude35SonnetProfile(LLMProfile):
    name: str = "anthropic-claude-3.5-sonnet"


@attrs.define(auto_attribs=True, frozen=True)
class AnthropicClaude37SonnetProfile(LLMProfile):
    name: str = "anthropic-claude-3.7-sonnet"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleChatBisonProfile(LLMProfile):
    name: str = "google-chat-bison"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleChatCodeBisonProfile(LLMProfile):
    name: str = "google-chat-code-bison"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleCompletionCodeGeckoProfile(LLMProfile):
    name: str = "google-completion-code-gecko"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleCompletionCodeBisonProfile(LLMProfile):
    name: str = "google-completion-code-bison"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleChatGeminiProProfile(LLMProfile):
    name: str = "google-chat-gemini-pro"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleChatGeminiUltraProfile(LLMProfile):
    name: str = "google-chat-gemini-ultra"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleChatGeminiPro15Profile(LLMProfile):
    name: str = "google-chat-gemini-pro-1.5"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleChatGeminiFlash15Profile(LLMProfile):
    name: str = "google-chat-gemini-flash-1.5"


@attrs.define(auto_attribs=True, frozen=True)
class GoogleChatGeminiPro25Profile(LLMProfile):
    name: str = "google-chat-gemini-pro-2.5"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIEmbeddingAdaProfile(LLMProfile):
    name: str = "openai-embedding-ada"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIEmbeddingSmallProfile(LLMProfile):
    name: str = "openai-embedding-small"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIEmbeddingLargeProfile(LLMProfile):
    name: str = "openai-embedding-large"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIChatGPTProfile(LLMProfile):
    name: str = "openai-chat-gpt"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIChatGPT16kProfile(LLMProfile):
    name: str = "openai-chat-gpt-16k"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIGPT4Profile(LLMProfile):
    name: str = "openai-gpt-4"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIGPT4TurboDeprecatedProfile(LLMProfile):
    name: str = "gpt-4-1106-preview"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIGPT4TurboProfile(LLMProfile):
    name: str = "openai-gpt-4-turbo"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIGPT4oDeprecatedProfile(LLMProfile):
    name: str = "gpt-4o"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIGPT4oProfile(LLMProfile):
    name: str = "openai-gpt-4o"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIGPT4OMiniProfile(LLMProfile):
    name: str = "openai-gpt-4o-mini"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIo1Profile(LLMProfile):
    name: str = "openai-o1"


@attrs.define(auto_attribs=True, frozen=True)
class OpenAIo1MiniProfile(LLMProfile):
    name: str = "openai-o1-mini"


class Profile:
    GRAZIE_GPT_NEO_TINY_TEXT = GrazieGptNeoTinyTextProfile()
    GRAZIE_REPLIT_CODE_V1_SMALL = GrazieReplitCodeV1SmallProfile()
    GRAZIE_BIGCODE_STARCODER = GrazieBigCodeStarcoderProfile()

    GRAZIE_CHAT_LLAMA_V2_7b = GrazieChatLlamaV2_7bProfile()
    GRAZIE_CHAT_LLAMA_V2_13b = GrazieChatLlamaV2_13bProfile()
    GRAZIE_CODE_LLAMA_7b = GrazieCodeLlama7bProfile()
    GRAZIE_CODE_LLAMA_13b = GrazieCodeLlama13bProfile()

    ANTHROPIC_CLAUDE = AnthropicClaudeProfile()
    ANTHROPIC_CLAUDE_INSTANT = AnthropicClaudeInstantProfile()

    ANTHROPIC_CLAUDE_3_HAIKU = AnthropicClaude3HaikuProfile()
    ANTHROPIC_CLAUDE_3_SONNET = AnthropicClaude3SonnetProfile()
    ANTHROPIC_CLAUDE_3_OPUS = AnthropicClaude3OpusProfile()
    ANTHROPIC_CLAUDE_35_SONNET = AnthropicClaude35SonnetProfile()
    ANTHROPIC_CLAUDE_37_SONNET = AnthropicClaude37SonnetProfile()

    GOOGLE_CHAT_BISON = GoogleChatBisonProfile()
    GOOGLE_CHAT_CODE_BISON = GoogleChatCodeBisonProfile()
    GOOGLE_COMPLETION_CODE_GECKO = GoogleCompletionCodeGeckoProfile()
    GOOGLE_COMPLETION_CODE_BISON = GoogleCompletionCodeBisonProfile()

    GOOGLE_CHAT_GEMINI_PRO = GoogleChatGeminiProProfile()
    GOOGLE_CHAT_GEMINI_ULTRA = GoogleChatGeminiUltraProfile()
    GOOGLE_CHAT_GEMINI_PRO_15 = GoogleChatGeminiPro15Profile()
    GOOGLE_CHAT_GEMINI_PRO_25 = GoogleChatGeminiPro25Profile()
    GOOGLE_CHAT_GEMINI_FLASH_15 = GoogleChatGeminiFlash15Profile()

    OPENAI_EMBEDDING_ADA = OpenAIEmbeddingAdaProfile()
    OPENAI_EMBEDDING_SMALL = OpenAIEmbeddingSmallProfile()
    OPENAI_EMBEDDING_LARGE = OpenAIEmbeddingLargeProfile()

    OPENAI_CHAT_GPT = OpenAIChatGPTProfile()
    OPENAI_CHAT_GPT_16k = OpenAIChatGPT16kProfile()
    OPENAI_GPT_4 = OpenAIGPT4Profile()
    OPENAI_GPT_4_TURBO_DEPRECATED = OpenAIGPT4TurboDeprecatedProfile()
    OPENAI_GPT_4_TURBO = OpenAIGPT4TurboProfile()
    OPENAI_GPT_4_O_DEPRECATED = OpenAIGPT4oDeprecatedProfile()
    OPENAI_GPT_4_O = OpenAIGPT4oProfile()
    OPENAI_GPT_4_O_MINI = OpenAIGPT4OMiniProfile()
    OPENAI_O_1 = OpenAIo1Profile()
    OPENAI_O_1_MINI = OpenAIo1MiniProfile()

    @classmethod
    def get_by_name(cls, name: str) -> LLMProfile:
        """Used to get profile string. Mainly to be able to pass profile as a string argument in argparse."""
        for attr in dir(Profile):
            value = getattr(Profile, attr)
            if isinstance(value, LLMProfile) and value.name == name:
                return value
        raise ValueError(f"Unknown profile name: {name}")



import attrs

from grazie.api.client.completion.prompt import CompletionPrompt


@attrs.define(auto_attribs=True)
class LLMCompleteEventV3:
    current: str


@attrs.define(auto_attribs=True, frozen=True)
class CompletionResponse:
    prompt: CompletionPrompt
    completion: str


@attrs.define(auto_attribs=True, frozen=True)
class CompletionResponseStream:
    chunk: str



from ._client import AsyncClient, Client


class BaseAPI:
    def __init__(self, client: Client):
        self.client = client


class BaseAsyncAPI:
    def __init__(self, client: AsyncClient):
        self.client = client



import json
from typing import Any, Dict, Optional

from .exceptions import SSERequestFailedError, SSEResponseParseError


def _decode_sse_line(line: str) -> Optional[Dict[str, Any]]:
    """
    The decoding algorithm is greatly simplified and ignores all messages
    that don't start with "data:". It also assumes, that all messages have
    the following format:
    ```
        data: {"event_type": "data|error", ...}
    ```
    """
    line = line.strip()

    if len(line) == 0:
        return None

    if not line.startswith("data:"):
        return None

    content = line[5:].strip()
    if content == "end":
        return None

    try:
        content_data = json.loads(content)
    except json.JSONDecodeError as e:
        raise SSEResponseParseError("SSE content is not a valid JSON") from e

    event_type = content_data.get("event_type")
    if event_type == "data":
        content_data.pop("event_type")
        return content_data
    elif event_type == "error":
        error_message = content_data.get("error_message", "unknown")
        error_details = f"Server error: {error_message!r}"
        raise SSERequestFailedError(error_details)
    else:
        raise SSEResponseParseError(
            f"Expected all data parts of sse response to be of event_type data or error, but got {event_type}. "
            "The client should be updated to the latest api version."
        )



import logging
import typing
import warnings
from typing import Any, AsyncIterator, Dict, Iterator, Optional

from .._api import BaseAPI, BaseAsyncAPI
from .types import (
    TaskCallResponse,
    TaskRoster,
    TaskStreamData,
    TaskStreamExecutionMetadata,
    TaskStreamFinishMetadata,
    TaskStreamFunctionCallMetadata,
    TaskStreamQuotaMetaData,
    TaskStreamText,
    TaskStreamUnknownMetadata,
)

logger = logging.getLogger(__name__)


class TasksAPI(BaseAPI):
    """TaskAPI client."""

    def execute_stream(self, id: str, parameters: dict) -> Iterator[TaskStreamData]:
        """Executes the task in TaskAPI.

        Args:
            id: task id from the roster
            parameters: task parameters from Swagger
        Returns:
            Stream of `TaskStreamData` objects in a form of iterator.
        """

        kwargs: typing.Dict[str, Any] = {}
        id, tag = split_id(id)
        if tag:
            kwargs["headers"] = {"Grazie-Task-Tag": tag}

        logger.debug(
            "Executing request to TaskAPI with task id %s and parameters: %r",
            f"{id}:{tag}" if tag else id,
            parameters,
        )
        for sse_event in self.client.stream(
            method="POST",
            path=f"task/stream/v4/{id}",
            json={"parameters": parameters},
            **kwargs,
        ):
            if event := _convert_sse_event(sse_event):
                yield event

    def execute(self, id: str, parameters: dict) -> TaskCallResponse:
        """Executes the task in TaskAPI.

        Args:
            id: task id from the roster
            parameters: task parameters from Swagger
        Returns:
            `TaskCallResponse` object that aggregates the streamed response from TaskAPI.
        """

        content = []
        quota_metadata = None
        execution_metadata = []
        finish_metadata = None
        unknown_metadata = None
        function_call_metadata = None

        for task_stream_data in self.execute_stream(id=id, parameters=parameters):
            if isinstance(task_stream_data, TaskStreamText):
                content.append(task_stream_data.content)
            elif isinstance(task_stream_data, TaskStreamQuotaMetaData):
                quota_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamExecutionMetadata):
                execution_metadata.append(task_stream_data)
            elif isinstance(task_stream_data, TaskStreamFinishMetadata):
                finish_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamUnknownMetadata):
                unknown_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamFunctionCallMetadata):
                function_call_metadata = task_stream_data

        return TaskCallResponse(
            content="".join(content),
            quota_metadata=quota_metadata,
            execution_metadata=execution_metadata,
            finish_metadata=finish_metadata,
            unknown_metadata=unknown_metadata,
            function_call_metadata=function_call_metadata,
        )

    def roster(self) -> TaskRoster:
        """
        Returns all available task ids.

        For the list of task parameters, see the Swagger documentation:
        https://api.app.stgn.grazie.aws.intellij.net/swagger-ui/index.html?urls.primaryName=Tasks#/
        """
        resp = self.client.request(
            method="GET",
            path="task/roster",
        )
        return TaskRoster.parse_raw(resp.read())


class AsyncTasksAPI(BaseAsyncAPI):
    """TaskAPI async client."""

    async def execute_stream(self, id: str, parameters: dict) -> AsyncIterator[TaskStreamData]:
        """Executes the task in TaskAPI.

        Args:
            id: task id from the roster
            parameters: task parameters from Swagger
        Returns:
            Stream of `TaskStreamData` objects in a form of async iterator.
        """

        kwargs: typing.Dict[str, Any] = {}
        id, tag = split_id(id)
        if tag:
            kwargs["headers"] = {"Grazie-Task-Tag": tag}

        logger.debug(
            "Executing request to TaskAPI with task id %s and parameters: %r",
            f"{id}:{tag}" if tag else id,
            parameters,
        )
        async for sse_event in self.client.stream(
            method="POST",
            path=f"task/stream/v4/{id}",
            json={"parameters": parameters},
            **kwargs,
        ):
            if event := _convert_sse_event(sse_event):
                yield event

    async def execute(self, id: str, parameters: dict) -> TaskCallResponse:
        """Executes the task in TaskAPI.

        Args:
            id: task id from the roster
            parameters: task parameters from Swagger
        Returns:
            `TaskCallResponse` object that aggregates the streamed response from TaskAPI.
        """

        content = []
        quota_metadata = None
        execution_metadata = []
        finish_metadata = None
        unknown_metadata = None
        function_call_metadata = None

        async for task_stream_data in self.execute_stream(id=id, parameters=parameters):
            if isinstance(task_stream_data, TaskStreamText):
                content.append(task_stream_data.content)
            elif isinstance(task_stream_data, TaskStreamQuotaMetaData):
                quota_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamExecutionMetadata):
                execution_metadata.append(task_stream_data)
            elif isinstance(task_stream_data, TaskStreamFinishMetadata):
                finish_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamUnknownMetadata):
                unknown_metadata = task_stream_data
            elif isinstance(task_stream_data, TaskStreamFunctionCallMetadata):
                function_call_metadata = task_stream_data

        return TaskCallResponse(
            content="".join(content),
            quota_metadata=quota_metadata,
            execution_metadata=execution_metadata,
            finish_metadata=finish_metadata,
            unknown_metadata=unknown_metadata,
            function_call_metadata=function_call_metadata,
        )

    async def roster(self) -> TaskRoster:
        """
        Returns all available task ids.

        For the list of task parameters, see the Swagger documentation:
        https://api.app.stgn.grazie.aws.intellij.net/swagger-ui/index.html?urls.primaryName=Tasks#/
        """
        resp = await self.client.request(
            method="GET",
            path="task/roster",
        )
        return TaskRoster.parse_raw(await resp.aread())


def _convert_sse_event(sse_event: Dict[str, Any]) -> Optional[TaskStreamData]:
    logger.debug("Received SSE from TaskAPI: %r", sse_event)

    event_type = sse_event.pop("type")

    if event_type == "Content":
        return TaskStreamText.parse_obj(sse_event)
    elif event_type == "QuotaMetadata":
        return TaskStreamQuotaMetaData.parse_obj(sse_event)
    elif event_type == "ExecutionMetadata":
        return TaskStreamExecutionMetadata.parse_obj(sse_event)
    elif event_type == "FinishMetadata":
        return TaskStreamFinishMetadata.parse_obj(sse_event)
    elif event_type == "UnknownMetadata":
        return TaskStreamUnknownMetadata.parse_obj(sse_event)
    elif event_type == "FunctionCallMetadata":
        return TaskStreamFunctionCallMetadata.parse_obj(sse_event)

    warnings.warn(f"Can't convert unknown task stream event type {event_type!r}")
    return None


def split_id(id: str) -> typing.Tuple[str, str]:
    if ":" in id:
        id, tag = id.split(":", 1)
        return id, tag
    return id, ""



from typing import Dict, FrozenSet

from grazie.api.client.parameters import Parameters


class LLMParameters:
    # Temperature of generation
    Temperature = Parameters.FloatKey("llm.parameters.temperature")

    # List of functions to use in generation
    Functions = Parameters.JsonKey("llm.parameters.functions")

    # Function call mode
    FunctionCall = Parameters.JsonKey("llm.parameters.function.mode")

    # Max length of generated text in tokens
    Length = Parameters.IntKey("llm.parameters.length")

    # Stop token to stop generation at. Token may contain more than one character.
    StopToken = Parameters.StrKey("llm.parameters.stop-token")

    # Restrict repetition of ngrams of no-repeat size
    NoRepeat = Parameters.IntKey("llm.parameters.no-repeat")

    # Nucleus sampling, where the model considers the results of the tokens with topP probability mass.
    TopP = Parameters.FloatKey("llm.parameters.top-p")

    # Only sample from the top K options for each subsequent token. Used to remove "long tail" low probability responses.
    TopK = Parameters.IntKey("llm.parameters.top-k")

    # The number of dimensions the resulting output embeddings should have
    Dimension = Parameters.IntKey("llm.parameters.dimension")

    # Controls the output mode to constrain the model to generate strings that parse into valid JSON objects or plain text.
    ResponseFormat = Parameters.JsonKey("llm.parameters.response-format")

    # Random number generator for the language model. Essential for ensuring reproducibility of outcomes.
    Seed = Parameters.IntKey("llm.parameters.seed")



from datetime import datetime
from decimal import Decimal
from enum import Enum
from typing import Optional

from pydantic import BaseModel


class AuthType(str, Enum):
    USER = "user"
    SERVICE = "service"
    APPLICATION = "application"

    @property
    def path(self) -> str:
        return {
            AuthType.USER: "/user/v5",
            AuthType.SERVICE: "/service/v5",
            AuthType.APPLICATION: "/application/v5",
        }[self]


class GatewayEndpoint(str, Enum):
    STAGING = "https://api.app.stgn.grazie.aws.intellij.net"
    PRODUCTION = "https://api.app.prod.grazie.aws.intellij.net"


def format_decimal(decimal: Decimal) -> str:
    formatted = str(decimal)

    if "." not in formatted:
        formatted = f"{formatted}."
    return formatted


class Credit(BaseModel):
    amount: Decimal

    class Config:
        json_encoders = {
            Decimal: format_decimal,
        }


class QuotaId(BaseModel):
    quotaId: str
    userId: Optional[str] = None


class Quota(BaseModel):
    """
    Represents a Quota - a set of data that encapsulates current usage
    of user and total allowed usage in Credit.

    Attributes:
        license: The license id the quota is for.
        current: The current usage of user in Credit.
        maximum: The maximum usage of user in Credit. Note that
            it may differ from the tariff, if extensions has been applied.
        until: The time after which the quota is considered expired, equals to license expiration time.
        quotaID: Used to identify quota
    """

    license: str
    current: Credit
    maximum: Credit
    until: Optional[datetime]
    quotaID: Optional[QuotaId]


class GrazieHeaders(str, Enum):
    """
    Headers that are used in grazie platform.
    """

    AUTH_TOKEN = "Grazie-Authenticate-JWT"
    ORIGINAL_USER_TOKEN = "Grazie-Original-User-JWT"
    ORIGINAL_APPLICATION_TOKEN = "Grazie-Original-Application-JWT"
    APPLICATION_USER = "Grazie-Application-User-ID"
    ORIGINAL_SERVICE_TOKEN = "Grazie-Original-Service-JWT"
    ORIGINAL_USER_IP = "Grazie-Original-User-IP"
    AGENT = "Grazie-Agent"
    QUOTA_METADATA = "Grazie-Quota-Metadata"
    TRACE_ID = "Grazie-Trace-Id"
    DEPRECATION_INFO = "Grazie-Deprecated-Info"
    ORIGINAL_USER_COUNTRY = "Grazie-Original-User-Country"



from typing import Any, List, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict

from ..types import Credit, Quota

TASK_TYPE_MAP = {
    str: "text",
    bool: "bool",
    int: "int",
    float: "double",
    bytes: "bytes",
    dict: "json",
}


class Key(BaseModel):
    fqdn: str
    type: str


class Value(BaseModel):
    type: str
    value: Any
    modified: Optional[int] = None


class Attributes(BaseModel):
    data: List[Union[Key, Value]]


class TaskStreamText(BaseModel):
    content: str


class TaskStreamQuotaMetaData(BaseModel):
    updated: Quota
    spent: Credit


class TaskStreamExecutionMetadata(BaseModel):
    attributes: Attributes


class TaskStreamFinishMetadata(BaseModel):
    reason: Literal["stop", "length", "function_call", "tool_call", "unknown"]


class TaskStreamUnknownMetadata(BaseModel):
    """
    Open model that can contain any arbitrary metadata fields
    """

    model_config = ConfigDict(extra="allow")


class TaskStreamFunctionCallMetadata(BaseModel):
    name: Optional[str] = None
    content: str


TaskStreamData = Union[
    TaskStreamText,
    TaskStreamQuotaMetaData,
    TaskStreamExecutionMetadata,
    TaskStreamFinishMetadata,
    TaskStreamUnknownMetadata,
    TaskStreamFunctionCallMetadata,
]


class TaskCallResponse(BaseModel):
    content: str
    quota_metadata: Optional[TaskStreamQuotaMetaData] = None
    execution_metadata: List[TaskStreamExecutionMetadata] = []
    finish_metadata: Optional[TaskStreamFinishMetadata] = None
    unknown_metadata: Optional[TaskStreamUnknownMetadata] = None
    function_call_metadata: Optional[TaskStreamFunctionCallMetadata] = None


class TaskRoster(BaseModel):
    ids: List[str]



from typing import Any, Dict, Union

import attrs


def filter_fields(obj: Dict[str, Any], t: Any):
    return {k: v for k, v in obj.items() if k in attrs.fields_dict(t)}


def typ(t: Any, **kwargs) -> Any:
    return attrs.field(converter=lambda obj: t(**filter_fields(obj, t)), **kwargs)


def typ_or_none(t: Any, **kwargs) -> Union[Any, None]:
    return attrs.field(
        converter=lambda obj: t(**filter_fields(obj, t)) if obj is not None else None,
        default=None,
        **kwargs,
    )



from typing import Dict, List

import attr

from grazie.api.client.chat.message import (
    AssistantChatMessage,
    ChatMessage,
    FunctionCallChatMessage,
    TextChatMessage,
)
from grazie.api.client.chat.roles import ChatRole


@attr.s(auto_attribs=True, frozen=False)
class ChatPrompt:
    messages: List[ChatMessage] = attr.ib(default=attr.Factory(list))

    def _add_text(self, item: str, role: ChatRole):
        self.messages.append(TextChatMessage(role, item))
        return self

    def add_user(self, item: str):
        return self._add_text(item, ChatRole.USER)

    def add_system(self, item: str):
        return self._add_text(item, ChatRole.SYSTEM)

    def add_assistant(self, item: str):
        self.messages.append(AssistantChatMessage(text=item))
        return self

    def add_assistant_function(self, name: str, content: str):
        self.messages.append(
            AssistantChatMessage(
                text=None,
                functionCall=AssistantChatMessage.FunctionCallResponse(
                    functionName=name, content=content
                ),
            )
        )
        return self

    def add_function(self, name: str, content: str):
        self.messages.append(FunctionCallChatMessage(name=name, content=content))
        return self

    def get_messages(self) -> List[Dict[str, str]]:
        return [message.to_json() for message in self.messages]



import json
import logging
import os
from enum import Enum
from typing import (
    Any,
    Dict,
    Iterable,
    Iterator,
    List,
    MutableMapping,
    Optional,
    Type,
    TypeVar,
)

import attrs
import cbor2 as cbor2
import requests
from requests.adapters import HTTPAdapter
from urllib3 import Retry

from grazie.api.client.chat.prompt import ChatPrompt
from grazie.api.client.chat.response import (
    ChatResponse,
    ChatResponseStream,
    LLMChatEventType,
    LLMChatEventV5,
    LLMChatEventV6,
)
from grazie.api.client.completion.prompt import CompletionPrompt
from grazie.api.client.completion.response import (
    CompletionResponse,
    CompletionResponseStream,
    LLMCompleteEventV3,
)
from grazie.api.client.emb.request import EmbeddingRequest, LLMEmbeddingRequest
from grazie.api.client.emb.response import EmbeddingResponse
from grazie.api.client.endpoints import GrazieApiGatewayUrls
from grazie.api.client.llm_parameters import LLMParameters
from grazie.api.client.parameters import Parameters
from grazie.api.client.profiles import LLMProfile, Profile
from grazie.api.client.qna.request import PrioritizedSource
from grazie.api.client.qna.response import AnswerStreamV2, RetrieveResponse
from grazie.api.client.quota.response import Quota

G_LOG = logging.getLogger(__name__)


@attrs.define(auto_attribs=True, frozen=True)
class GrazieAgent:
    name: str
    version: str


class GrazieHeaders(str, Enum):
    AUTH_TOKEN = "Grazie-Authenticate-JWT"
    ORIGINAL_USER_TOKEN = "Grazie-Original-User-JWT"
    ORIGINAL_APPLICATION_TOKEN = "Grazie-Original-Application-JWT"
    ORIGINAL_SERVICE_TOKEN = "Grazie-Original-Service-JWT"
    ORIGINAL_USER_IP = "Grazie-Original-User-IP"
    AGENT = "Grazie-Agent"
    QUOTA_METADATA = "Grazie-Quota-Metadata"


class AuthType(str, Enum):
    SERVICE = "service"
    USER = "user"
    APPLICATION = "application"


class SseResponseParseException(Exception):
    """Exception occurring during parsing of sse response.

    Can occur, for example, if api gets changed without client update."""


class RequestFailedException(Exception):
    """Exception which occurs when we receive non 200 response code from api gateway."""


class GrazieApiGatewayClient:
    def __init__(
        self,
        grazie_agent: GrazieAgent = GrazieAgent(name="default", version="dev"),
        url: str = GrazieApiGatewayUrls.PRODUCTION,
        auth_type: AuthType = AuthType.SERVICE,
        grazie_jwt_token: Optional[str] = None,
    ):
        """Creates and instance of a gateway client.

        Parameters
        ----------
        grazie_agent: meaningful identifier of the application or user which is using the api.
          It can later be used to calculate statistics on the server side, so make sure to set the correct version and update it accordingly.
          If you don't care about statistics, just use "dev" as a version, i.e. GrazieAgent(name="llm-eval", version="dev")
        url: address of the service, default points to staging, to get prod one, replace stgn with prod in url.
        auth_type: authorization type which should be set to service if you write a service, it's fine to use user for local testing and evaluation.
        grazie_jwt_token: JWT token which is used for authorization, user token can be obtained with Grazie Playground (try.ai.intellij.net).
          If you write an application and need service auth_type, contact Vladislav Tankov and he will issue the service token.
        """
        if grazie_jwt_token is None and "GRAZIE_JWT_TOKEN" not in os.environ:
            if "GRAZIE_USER_JWT_TOKEN" in os.environ:
                auth_type = AuthType.USER
            elif "GRAZIE_SERVICE_JWT_TOKEN" in os.environ:
                auth_type = AuthType.SERVICE
            else:
                raise ValueError(
                    """
                    Cannot set jwt token. Either pass it as a constructor param or set GRAZIE_JWT_TOKEN environment variable.
                    You can obtain user jwt token by going to https://play.stgn.grazie.ai/ and copying it by pressing a button in a top left corner.
                """
                )

        self._api_gateway_url = url
        self._auth_type = auth_type
        self._grazie_jwt_token = (
            grazie_jwt_token
            or os.environ.get("GRAZIE_JWT_TOKEN")
            or os.environ.get("GRAZIE_USER_JWT_TOKEN")
            or os.environ["GRAZIE_SERVICE_JWT_TOKEN"]
        )
        self._grazie_agent = json.dumps(attrs.asdict(grazie_agent))

    def _generate_headers(self, user_headers: Optional[Dict[str, str]] = None):
        return {
            GrazieHeaders.AUTH_TOKEN: self._grazie_jwt_token,
            GrazieHeaders.AGENT: self._grazie_agent,
            "Content-Type": "application/json",
            **(user_headers if user_headers else {}),
        }

    def complete(
        self,
        prompt: CompletionPrompt,
        profile: LLMProfile,
        parameters: Optional[Dict[Parameters.Key, Parameters.Value]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> CompletionResponse:
        chunks: List[str] = []
        for stream_response in self.complete_stream(
            prompt, profile=profile, parameters=parameters, headers=headers
        ):
            chunks.append(stream_response.chunk)

        return CompletionResponse(prompt=prompt, completion="".join(chunks))

    def complete_stream(
        self,
        prompt: CompletionPrompt,
        profile: LLMProfile,
        parameters: Optional[Dict[Parameters.Key, Parameters.Value]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Iterator[CompletionResponseStream]:
        request_data: Dict[str, Any] = {
            "prompt": prompt.message,
            "profile": profile.name,
            "postfix": prompt.suffix,
        }

        if parameters:
            request_data["parameters"] = self._create_parameters_json(parameters, profile)

        response = requests.post(
            f"{self._api_gateway_url}/{self._auth_type.value}/v5/llm/complete/stream/v3",
            headers=self._generate_headers(headers),
            json=request_data,
            timeout=300,
            stream=True,
        )

        self._raise_if_error(response)

        for chunk in _decode_llm_sse(
            self._parse_sse_response(response.iter_lines()), LLMCompleteEventV3
        ):
            yield CompletionResponseStream(chunk=chunk.current)

    def chat(
        self,
        chat: ChatPrompt,
        profile: LLMProfile,
        prompt_id: str = "unknown",
        parameters: Optional[Dict[Parameters.Key, Parameters.Value]] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> ChatResponse:
        """
        Calls chat api endpoint

        :param chat: chat history, usually ending with user message, some models also support system messages
        :param profile: which model to use, i.e., chatGPT, Vicuna, ...
        :param prompt_id: id of your prompt or feature which can be used to track expenses
        :param parameters: parameters to use during generation, supported parameters are located in LLMParameters
        Example:
            self.chat(
                chat=ChatPrompt()
                    .add_system("You are a helpful assistant.")
                    .add_user("Who won the world series in 2020?"),
                profile=Profile.OPENAI_CHAT_GPT,
                prompt_id="my-world-series-app",
                parameters={
                    LLMParameters.Temperature: Parameters.FloatValue(0.0)
                }
            )
        """
        function_call = None
        updated = None
        spent = None

        chunks: List[ChatResponseStream] = []
        for stream_response in self.chat_stream(
            chat,
            profile=profile,
            prompt_id=prompt_id,
            parameters=parameters,
            headers=headers,
        ):
            chunks.append(stream_response)
            if stream_response.function_call:
                function_call = stream_response.function_call
            if stream_response.updated:
                updated = stream_response.updated
            if stream_response.spent:
                spent = stream_response.spent

        return ChatResponse(
            prompt=chat,
            content="".join(c.chunk for c in chunks),
            function_call=function_call,
            updated=updated,
            spent=spent,
        )

    def chat_stream(
        self,
        chat: ChatPrompt,
        profile: LLMProfile,
        prompt_id: str = "unknown",
        parameters: Optional[Dict[Parameters.Key, Parameters.Value]] = None,
        headers: Optional[Dict[str, str]] = None,
        proxies: Optional[MutableMapping[str, str]] = None,
    ) -> Iterator[ChatResponseStream]:
        """
        Calls chat api endpoint in a streaming mode.

        :param chat: chat history, usually ending with user message, some models also support system messages
        :param profile: which model to use, i.e., chatGPT, Vicuna, ...
        :param prompt_id: id of your prompt or feature which can be used to track expenses
        :param parameters: parameters to use during generation, supported parameters are located in LLMParameters
        Example:
            self.chat_stream(
                chat=ChatPrompt()
                    .add_system("You are a helpful assistant.")
                    .add_user("Who won the world series in 2020?"),
                profile=Profile.OPENAI_CHAT_GPT,
                prompt_id="my-world-series-app",
                parameters={
                    LLMParameters.Temperature: Parameters.FloatValue(0.0)
                }
            )
        """
        request_data: Dict[str, Any] = {
            "chat": {"messages": chat.get_messages()},
            "prompt": prompt_id,
            "profile": profile.name,
        }

        if parameters:
            request_data["parameters"] = self._create_parameters_json(parameters, profile)

        response = requests.post(
            url=f"{self._api_gateway_url}/{self._auth_type.value}/v5/llm/chat/stream/v6",
            headers=self._generate_headers(headers),
            json=request_data,
            timeout=300,
            stream=True,
            proxies=proxies,
        )

        self._raise_if_error(response)

        for event in _decode_llm_sse(
            self._parse_sse_response(response.iter_lines()), LLMChatEventV6
        ):
            event: LLMChatEventV6

            if event.type is LLMChatEventType.Content:
                yield ChatResponseStream(chunk=event.content)
            elif event.type is LLMChatEventType.FunctionCall:
                yield ChatResponseStream(chunk=event.content, function_call=event.name)
            elif event.type is LLMChatEventType.QuotaMetadata:
                yield ChatResponseStream(chunk="", spent=event.spent, updated=event.updated)
            else:
                raise ValueError(f"Unsupported event type {event.type!r}")

    def quota(self, headers: Optional[Dict[str, str]] = None) -> Dict[str, Quota]:
        with requests.Session() as s:
            s.mount(
                f"{self._api_gateway_url}",
                HTTPAdapter(
                    max_retries=Retry(
                        3,
                        backoff_factor=0.1,
                        status_forcelist=[429, 500, 502, 503, 504],
                    )
                ),
            )
            response = requests.post(
                f"{self._api_gateway_url}/{self._auth_type.value}/v5/quota/get",
                headers=self._generate_headers(headers),
                timeout=300,
            )

        self._raise_if_error(response)
        return {q_name: Quota(**q) for q_name, q in response.json().items()}

    def embed(
        self, request: EmbeddingRequest, headers: Optional[Dict[str, str]] = None
    ) -> EmbeddingResponse:
        request_options = []
        if request.format_cbor:
            request_options.append("format=cbor")
        if request.normalize:
            request_options.append("normalize=true")
        request_data: Dict[str, Any] = {
            "texts": request.texts,
            "options": request_options,
        }
        if request.model:
            request_data["model"] = request.model
        with requests.Session() as s:
            s.mount(
                f"{self._api_gateway_url}",
                HTTPAdapter(
                    max_retries=Retry(
                        3,
                        backoff_factor=0.1,
                        status_forcelist=[429, 500, 502, 503, 504],
                    )
                ),
            )
            response = requests.post(
                f"{self._api_gateway_url}/{self._auth_type.value}/v5/meta/emb/embed",
                headers=self._generate_headers(headers),
                json=request_data,
                timeout=300,
            )

        self._raise_if_error(response)

        body: Dict
        if response.headers["Content-Type"] == "application/cbor":
            body = cbor2.loads(response.content)
        else:
            body = json.loads(response.content)
        embeddings = [e["values"] for e in body["embeddings"]]
        return EmbeddingResponse(embeddings=embeddings)

    def llm_embed(
        self, request: LLMEmbeddingRequest, headers: Optional[Dict[str, str]] = None
    ) -> EmbeddingResponse:
        request_data: Dict[str, Any] = {"texts": request.texts, "profile": request.profile.name}

        if request.dimensions:
            parameters: Dict[Parameters.Key, Parameters.Value] = {
                LLMParameters.Dimension: Parameters.IntValue(request.dimensions)
            }
            request_data["parameters"] = self._create_parameters_json(parameters, request.profile)

        with requests.Session() as s:
            s.mount(
                f"{self._api_gateway_url}",
                HTTPAdapter(
                    max_retries=Retry(
                        3, backoff_factor=0.1, status_forcelist=[429, 500, 502, 503, 504]
                    )
                ),
            )
            response = requests.post(
                f"{self._api_gateway_url}/{self._auth_type.value}/v5/llm/embedding/v2",
                headers=self._generate_headers(headers),
                json=request_data,
                timeout=300,
            )

        self._raise_if_error(response)

        body: Dict[str, List[Dict]] = cbor2.loads(response.content)
        embeddings = [e["values"] for e in body["embeddings"]]

        return EmbeddingResponse(embeddings=embeddings)

    def retrieve(
        self,
        query: str,
        data_source: str,
        profile: Optional[LLMProfile] = None,
        size: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> RetrieveResponse:
        request_data: Dict[str, Any] = {
            "query": query,
            "dataSource": data_source,
        }
        if size is not None:
            request_data["size"] = size
        if profile is not None:
            request_data["llmProfile"] = profile.name

        with requests.Session() as s:
            s.mount(
                f"{self._api_gateway_url}",
                HTTPAdapter(
                    max_retries=Retry(
                        3, backoff_factor=0.1, status_forcelist=[429, 500, 502, 503, 504]
                    )
                ),
            )
            response = requests.post(
                url=f"{self._api_gateway_url}/{self._auth_type.value}/v5/meta/qa/retrieve/v1",
                headers=self._generate_headers(headers),
                json=request_data,
                timeout=300,
                stream=True,
            )

        self._raise_if_error(response)

        body: Dict[str, List[Dict]] = json.loads(response.content)
        return RetrieveResponse(documents=body["documents"])

    def retrieve_v2(
        self,
        query: str,
        config_name: str,
        data_source_lists: List[List[PrioritizedSource]],
        profile: Optional[LLMProfile] = None,
        size: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> RetrieveResponse:
        request_data: Dict[str, Any] = {
            "query": query,
            "config": config_name,
            "prioritizedSources": [
                [attrs.asdict(source) for source in source_list]
                for source_list in data_source_lists
            ],
        }
        if size is not None:
            request_data["size"] = size
        if profile is not None:
            request_data["llmProfile"] = profile.name

        with requests.Session() as s:
            s.mount(
                f"{self._api_gateway_url}",
                HTTPAdapter(
                    max_retries=Retry(
                        3, backoff_factor=0.1, status_forcelist=[429, 500, 502, 503, 504]
                    )
                ),
            )
            response = requests.post(
                url=f"{self._api_gateway_url}/{self._auth_type.value}/v5/meta/qa/retrieve/v2",
                headers=self._generate_headers(headers),
                json=request_data,
                timeout=300,
                stream=False,
            )

        self._raise_if_error(response)

        body: Dict[str, List[Dict]] = json.loads(response.content)
        return RetrieveResponse(documents=body["documents"])

    def answer_stream(
        self,
        query: str,
        data_source: str,
        profile: LLMProfile = Profile.OPENAI_GPT_4_TURBO,
        size: Optional[int] = None,
        headers: Optional[Dict[str, str]] = None,
    ) -> Iterator[AnswerStreamV2]:
        request_data: Dict[str, Any] = {
            "query": query,
            "dataSource": data_source,
            "llmProfile": profile.name,
        }
        if size is not None:
            request_data["size"] = size

        response = requests.post(
            url=f"{self._api_gateway_url}/{self._auth_type.value}/v5/meta/qa/answer/v2",
            headers=self._generate_headers(headers),
            json=request_data,
            timeout=300,
            stream=True,
        )

        yield from _decode_llm_sse(self._parse_sse_response(response.iter_lines()), AnswerStreamV2)

    def _raise_if_error(self, response: requests.Response):
        if response.status_code != 200:
            raise RequestFailedException(
                f"{response.status_code} Error: {response.reason}. {response.text}"
            )

    @staticmethod
    def _create_parameters_json(
        parameters: Dict[Parameters.Key, Parameters.Value], llm_profile: LLMProfile
    ) -> Dict[str, Any]:
        return {"data": Parameters.serialize(parameters)}

    @staticmethod
    def _parse_sse_response(lines: Iterable[bytes]) -> Iterable[Dict[str, Any]]:
        for line in lines:
            line = line.decode()
            if len(line.strip()) == 0:
                continue

            if not line.startswith("data:"):
                raise SseResponseParseException(
                    "Got non-data part in sse response. The client should be updated to the latest api version."
                )

            content = line[5:].strip()
            if content == "end":
                return

            content_data = json.loads(content)

            event_type = content_data["event_type"]
            if event_type == "data":
                content_data.pop("event_type")
                yield content_data
            elif event_type == "error":
                error_message = f"Server error"
                if "error_message" in content_data:
                    error_message += error_message + ": " + content_data["error_message"]
                raise RequestFailedException(error_message)
            else:
                raise SseResponseParseException(
                    f"Expected all data parts of sse response to be of event_type data or error, but got {event_type}. "
                    "The client should be updated to the latest api version."
                )

        raise SseResponseParseException(
            'Expected "data: end" at the end of response, but not found.'
        )


T = TypeVar("T", LLMChatEventV6, LLMChatEventV5, LLMCompleteEventV3, AnswerStreamV2)


def _decode_llm_sse(events: Iterable[Dict[str, Any]], response_type: Type[T]) -> Iterable[T]:
    yield from (response_type(**event) for event in events)
